@@ -134,12 +134,27 @@ selection_changed (AdwComboRow *self)
 {
   AdwComboRowPrivate *priv = adw_combo_row_get_instance_private (self);
   guint selected;
+  GtkFilter *filter;
 
   if (!GTK_IS_SINGLE_SELECTION (priv->selection))
     return;
 
   selected = gtk_single_selection_get_selected (GTK_SINGLE_SELECTION (priv->selection));
 
+  /* reset the filter so positions are 1-1 */
+  filter = gtk_filter_list_model_get_filter (GTK_FILTER_LIST_MODEL (priv->filter_model));
+  if (GTK_IS_STRING_FILTER (filter))
+    gtk_string_filter_set_search (GTK_STRING_FILTER (filter), "");
+  gtk_single_selection_set_selected (GTK_SINGLE_SELECTION (priv->popup_selection), selected);
+
+  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_SELECTED]);
+}
+
+static void
+selection_item_changed (AdwComboRow *self)
+{
+  AdwComboRowPrivate *priv = adw_combo_row_get_instance_private (self);
+
   if (priv->use_subtitle) {
     if (g_list_model_get_n_items (G_LIST_MODEL (priv->current_selection)) > 0) {
       GtkListItem *item = g_list_model_get_item (G_LIST_MODEL (priv->current_selection), 0);
@@ -154,9 +169,6 @@ selection_changed (AdwComboRow *self)
     }
   }
 
-  gtk_single_selection_set_selected (GTK_SINGLE_SELECTION (priv->popup_selection), selected);
-
-  g_object_notify_by_pspec (G_OBJECT (self), props[PROP_SELECTED]);
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_SELECTED_ITEM]);
 }
 
@@ -175,9 +187,14 @@ static void
 row_activated_cb (AdwComboRow *self)
 {
   AdwComboRowPrivate *priv = adw_combo_row_get_instance_private (self);
+  GtkFilter *filter;
 
   gtk_popover_popdown (GTK_POPOVER (priv->popover));
 
+  /* reset the filter so positions are 1-1 */
+  filter = gtk_filter_list_model_get_filter (GTK_FILTER_LIST_MODEL (priv->filter_model));
+  if (GTK_IS_STRING_FILTER (filter))
+    gtk_string_filter_set_search (GTK_STRING_FILTER (filter), "");
   adw_combo_row_set_selected (self, gtk_single_selection_get_selected (GTK_SINGLE_SELECTION (priv->popup_selection)));
 }
 
@@ -477,6 +494,7 @@ adw_combo_row_dispose (GObject *object)
 
   if (priv->selection) {
     g_signal_handlers_disconnect_by_func (priv->selection, selection_changed, self);
+    g_signal_handlers_disconnect_by_func (priv->selection, selection_item_changed, self);
     g_signal_handlers_disconnect_by_func (priv->selection, model_changed, self);
   }
 
@@ -807,6 +825,7 @@ adw_combo_row_set_model (AdwComboRow *self,
 
     if (priv->selection) {
       g_signal_handlers_disconnect_by_func (priv->selection, selection_changed, self);
+      g_signal_handlers_disconnect_by_func (priv->selection, selection_item_changed, self);
       g_signal_handlers_disconnect_by_func (priv->selection, model_changed, self);
     }
 
@@ -828,7 +847,7 @@ adw_combo_row_set_model (AdwComboRow *self,
     gtk_list_view_set_model (priv->list, selection);
     g_object_unref (selection);
 
-    selection = GTK_SELECTION_MODEL (gtk_single_selection_new (g_object_ref (filter_model)));
+    selection = GTK_SELECTION_MODEL (gtk_single_selection_new (g_object_ref (model)));
     g_set_object (&priv->selection, selection);
     g_object_unref (selection);
 
@@ -841,9 +860,11 @@ adw_combo_row_set_model (AdwComboRow *self,
     g_object_unref (selection);
 
     g_signal_connect_swapped (priv->selection, "notify::selected", G_CALLBACK (selection_changed), self);
+    g_signal_connect_swapped (priv->selection, "notify::selected-item", G_CALLBACK (selection_item_changed), self);
     g_signal_connect_swapped (priv->selection, "items-changed", G_CALLBACK (model_changed), self);
 
     selection_changed (self);
+    selection_item_changed (self);
     model_changed (self);
   }
 
