@@ -247,6 +247,7 @@ static void
 close_idle_cb (AdwTab *self)
 {
   adw_tab_view_close_page (self->view, self->page);
+  g_object_unref (self);
 }
 
 static void
@@ -258,7 +259,7 @@ close_clicked_cb (AdwTab *self)
   /* When animations are disabled, we don't want to immediately remove the
    * whole tab mid-click. Instead, defer it until the click has happened.
    */
-  g_idle_add_once ((GSourceOnceFunc) close_idle_cb, self);
+  g_idle_add_once ((GSourceOnceFunc) close_idle_cb, g_object_ref (self));
 }
 
 static void
